#!/usr/bin/env python3

"""
Use ``debug-dump-search-scores.py`` to dump all information to some HDF file.
The you can use this script here just for visualization/plotting.
"""

from __future__ import annotations

from math import ceil
import typing
import os
from pprint import pprint
import argparse
import h5py
import numpy
from glob import glob

import _setup_returnn_env  # noqa
from returnn.datasets.lm import Lexicon
import returnn.util.basic as util


class Alignment:
    """Phone/word alignment"""

    class Item:
        """Phone or word"""

        def __init__(self, value, start_frame, end_frame=None, word_start=True, word_end=True):
            """
            :param str value:
            :param int start_frame:
            :param int|None end_frame:
            :param bool word_start:
            :param bool word_end:
            """
            self.value = value
            self.word_start = word_start
            self.word_end = word_end
            self.start_frame = start_frame
            if end_frame is None:
                end_frame = self.start_frame
            self.end_frame = end_frame

        @property
        def num_frames(self):
            return self.end_frame - self.start_frame + 1

        def copy(self):
            """
            :rtype: Alignment.Item
            """
            return Alignment.Item(
                value=self.value,
                start_frame=self.start_frame,
                end_frame=self.end_frame,
                word_start=self.word_start,
                word_end=self.word_end,
            )

        __repr__ = util.simple_obj_repr

        def __eq__(self, other):
            """
            :param Alignment.Item other:
            :rtype: bool
            """
            return repr(self) == repr(other)

        def __ne__(self, other):
            return not (self == other)

    def __init__(self):
        self.items = []  # type: typing.List[Alignment.Item]

    def __repr__(self):
        return "Alignment{%s}" % ", ".join(
            ["%r %i-%i" % (item.value, item.start_frame, item.end_frame) for item in self.items]
        )

    def __eq__(self, other):
        """
        :param Alignment other:
        :rtype: bool
        """
        return self.items == other.items

    def __ne__(self, other):
        return not (self == other)

    def copy(self):
        """
        :rtype: Alignment
        """
        res = Alignment()
        res.items = [item.copy() for item in self.items]
        return res

    def copy_extended(self, new_item):
        """
        :param Alignment.Item new_item:
        :rtype: Alignment
        """
        res = self.copy()
        res.items.append(new_item)
        return res


def parse_phone_alignment(filename):
    """
    :param str filename:
    :rtype: Alignment
    """
    lines = open(filename, "r").read().splitlines()
    # Example:
    """
  <?xml version="1.0" encoding="ISO-8859-1"?>
  <sprint>
    time= 0       emission=       10989   allophone=      ow{#+k}@i       index=  10989   state=  0
    time= 1       emission=       10989   allophone=      ow{#+k}@i       index=  10989   state=  0
    time= 2       emission=       10989   allophone=      ow{#+k}@i       index=  10989   state=  0
  """
    import re

    res = Alignment()
    line_re = re.compile(
        "^time=\\s*([0-9]+)\\s*"
        "emission=\\s*([0-9]+)\\s*"
        "allophone=\\s*([A-Za-z{}\\[\\]+#@]+)\\s*"
        "index=\\s*([0-9]+)\\s*"
        "state=\\s*([0-9]+)$"
    )
    allophone_re = re.compile("^([A-Za-z\\[\\]]+){([#a-z]+)\\+([#a-z]+)}([@a-z]*)$")
    time_i = 0
    last_state_i = -1
    last_flags_s = None
    for line in lines:
        line = line.strip()
        if not line or line.startswith("<"):
            continue
        line_match = line_re.match(line)
        assert line_match, "line: %r" % line
        time_s, emission_s, allophone_s, index_s, state_s = line_match.groups()
        state_i = int(state_s)
        assert int(time_s) == time_i
        allophone_match = allophone_re.match(allophone_s)
        assert allophone_match, "allophone: %r" % allophone_s
        phone_s, phone_left_s, phone_right_s, phone_flags_s = allophone_match.groups()
        if res.items and res.items[-1].value == phone_s and state_i >= last_state_i and phone_flags_s == last_flags_s:
            res.items[-1].end_frame += 1
        else:
            res.items.append(
                Alignment.Item(
                    value=phone_s, word_start="@i" in phone_flags_s, word_end="@f" in phone_flags_s, start_frame=time_i
                )
            )
        last_state_i = state_i
        last_flags_s = phone_flags_s
        time_i += 1
    assert sum([p.num_frames for p in res.items]) == time_i
    return res


def parse_phone_alignments(pattern):
    """
    :param str pattern: e.g. 'align-*.txt'. stdout from RASR archiver
    :return: name -> list of phones
    :rtype: dict[str,Alignment]
    """
    res = {}
    assert pattern.count("*") == 1
    prefix, postfix = pattern.split("*")
    for fn in glob(pattern):
        assert fn.startswith(prefix) and fn.endswith(postfix)
        name = fn[len(prefix) : -len(postfix)]
        res[name] = parse_phone_alignment(fn)
    return res


def select_phone_alignment(phone_alignments, seq_tag):
    """
    :param dict[str,Alignment] phone_alignments:
    :param str seq_tag:
    :rtype: Alignment
    """
    if seq_tag in phone_alignments:
        phone_alignment = phone_alignments[seq_tag]
    elif os.path.basename(seq_tag) in phone_alignments:
        phone_alignment = phone_alignments[os.path.basename(seq_tag)]
    else:
        raise Exception(
            "Did not find phone alignment for seq tag %r. Available: %r" % (seq_tag, list(phone_alignments.keys()))
        )
    return phone_alignment


def phone_alignment_to_word_alignment(lexicon, words, phone_alignment):
    """
    :param Lexicon lexicon:
    :param list[str] words:
    :param Alignment phone_alignment:
    :return: list of words, and number of time frames (time aligned)
    :rtype: Alignment|None
    """
    sil_phone = "[SILENCE]"
    num_non_silence_frames = sum([p.num_frames for p in phone_alignment.items if p.value != sil_phone])

    class SingleState:
        """Single state"""

        def __init__(self, word_i, lemma_phones=None, lemma_phone_i=-1, word_alignment=None):
            """
            :param int word_i:
            :param list[str]|None lemma_phones:
            :param int lemma_phone_i:
            :param Alignment word_alignment:
            """
            if not word_alignment:
                word_alignment = Alignment()
            self.word_i = word_i
            self.lemma_phones = lemma_phones
            self.lemma_phone_i = lemma_phone_i
            self.word_alignment = word_alignment

        __repr__ = util.simple_obj_repr

        def is_lemma_finished(self):
            if not self.lemma_phones:
                return True
            if self.lemma_phone_i >= len(self.lemma_phones):
                return True
            return False

        def get_with_new_lemma(self, phone):
            """
            :param Alignment.Item phone:
            :rtype: typing.Generator[SingleState]
            """
            assert self.is_lemma_finished()
            if not phone.word_start:
                return
            if phone.value == sil_phone:
                yield self
                return
            word_i = self.word_i + 1
            if word_i >= len(words):
                return
            assert word_i < len(words)
            word = words[word_i]
            new_word_alignment = self.word_alignment.copy_extended(
                Alignment.Item(value=word, start_frame=phone.start_frame, end_frame=phone.end_frame)
            )
            lemma = lexicon.lemmas[word]
            lemma_phones_opts = lemma["phons"]
            assert isinstance(lemma_phones_opts, list)
            for lemma_phones in lemma_phones_opts:
                lemma_phones = lemma_phones["phon"]
                assert isinstance(lemma_phones, str) and len(lemma_phones) > 0
                lemma_phones = lemma_phones.split()
                assert isinstance(lemma_phones, list) and len(lemma_phones) > 0 and isinstance(lemma_phones[0], str)
                if lemma_phones[0] == phone.value:
                    yield SingleState(
                        word_i=word_i,
                        lemma_phones=lemma_phones,
                        lemma_phone_i=1,
                        word_alignment=new_word_alignment.copy(),
                    )

        def get_inc_phone_in_lemma(self, phone):
            """
            :param Alignment.Item phone:
            :rtype: typing.Generator[SingleState]
            """
            if self.lemma_phones[self.lemma_phone_i] != phone.value:
                return
            if self.lemma_phone_i == len(self.lemma_phones) - 1:
                if not phone.word_end:
                    return
            new_word_alignment = self.word_alignment.copy()
            new_word_alignment.items[-1].end_frame += phone.num_frames
            yield SingleState(
                word_i=self.word_i,
                lemma_phones=self.lemma_phones,
                lemma_phone_i=self.lemma_phone_i + 1,
                word_alignment=new_word_alignment,
            )

    possible_states = [SingleState(word_i=-1, lemma_phones=None, lemma_phone_i=-1, word_alignment=Alignment())]

    for phone in phone_alignment.items:
        next_possible_states = []  # type: typing.List[SingleState]
        for state in possible_states:
            if state.is_lemma_finished():
                next_possible_states.extend(state.get_with_new_lemma(phone=phone))
            else:
                next_possible_states.extend(state.get_inc_phone_in_lemma(phone=phone))

        if not next_possible_states:
            break
        possible_states = next_possible_states

    final_possible_states = []  # type: typing.List[SingleState]
    for state in possible_states:
        if state.is_lemma_finished() and state.word_i == len(words) - 1:
            final_possible_states.append(state)

    if not final_possible_states:
        return None
    assert len(final_possible_states) == 1
    word_alignment = final_possible_states[0].word_alignment
    assert sum([w.num_frames for w in word_alignment.items]) == num_non_silence_frames
    return word_alignment


class AlignmentPlotter:
    """
    Using matplotlib.
    Can be hard or soft alignments.
    """

    def __init__(self):
        self.recent_seq_tag = None  # type: typing.Optional[str]
        self.plot_by_seq_tag = {}  # type: typing.Dict[str,AlignmentPlotter.Plot]

    def set_recent_seq_tag(self, seq_tag):
        """
        :param str seq_tag:
        """
        self.recent_seq_tag = seq_tag
        if seq_tag in self.plot_by_seq_tag:
            return
        self.plot_by_seq_tag[seq_tag] = AlignmentPlotter.Plot(seq_tag=seq_tag, parent=self)

    @property
    def recent_plot(self):
        return self.plot_by_seq_tag[self.recent_seq_tag]

    class Alignment:
        """
        (Soft) alignment to plot, e.g. attention weights.
        """

        def __init__(
            self,
            plot,
            key,
            human,
            alignment,
            time_reduction=1,
            add_extra_time=0,
            add_extra_time_red=0,
            cmap=None,
            cmap_palette=None,
            colorbar=False,
        ):
            """
            :param AlignmentPlotter plot:
            :param str key:
            :param str human:
            :param list[int]|numpy.ndarray|Alignment alignment:
            :param int time_reduction:
            :param int add_extra_time:
            :param int add_extra_time_red:
            :param cmap:
            :param int|None cmap_palette:
            :param bool colorbar:
            """
            self.plot = plot
            self.seq_tag = plot.recent_seq_tag
            self.key = key
            self.colorbar = colorbar
            self.human = human
            self.cmap = cmap
            self.cmap_palette = cmap_palette
            self.alignment = alignment
            self.time_reduction = time_reduction
            self.add_extra_time = add_extra_time
            self.add_extra_time_red = add_extra_time_red
            self.enabled = True
            self.mapped_alignment = None  # type: typing.Optional[numpy.ndarray]

        def __repr__(self):
            return "AlignmentPlotter.Alignment(human=%r, key=%r, seq_tag=%r)" % (self.human, self.key, self.seq_tag)

        def add_to_plot(self, overwrite=False):
            assert self.plot.recent_plot.seq_tag == self.seq_tag
            if not overwrite:
                assert self.key not in self.plot.recent_plot.alignments
            self.plot.recent_plot.alignments[self.key] = self

        def _check_alignment(self):
            """
            :return: whether to use this
            :rtype: bool
            """
            plot = self.plot.plot_by_seq_tag[self.seq_tag]  # recent_plot might be different
            time_length = plot.time_length + self.add_extra_time
            if isinstance(self.alignment, numpy.ndarray):
                assert self.alignment.shape == (
                    len(plot.output),
                    ceil(time_length / self.time_reduction) + self.add_extra_time_red,
                ), "%s: output len %i, orig len %i, time red %s, extra %i, extra red %i, align shape %r" % (
                    self,
                    len(plot.output),
                    plot.time_length,
                    self.time_reduction,
                    self.add_extra_time,
                    self.add_extra_time_red,
                    self.alignment.shape,
                )
                if self.time_reduction != 1:
                    assert isinstance(self.time_reduction, int) and self.time_reduction > 1  # else not implemented...
                    new_align = numpy.repeat(self.alignment, self.time_reduction, axis=1)
                    assert new_align.shape[1] >= time_length
                    self.alignment = new_align
            elif isinstance(self.alignment, list):
                assert len(self.alignment) == len(plot.output)
                alignment = numpy.zeros(
                    (len(plot.output), ceil(time_length / self.time_reduction) + self.add_extra_time_red),
                    dtype="float32",
                )
                for i, j in enumerate(self.alignment):
                    alignment[i, j] = 1.0
                alignment = numpy.repeat(alignment, self.time_reduction, axis=1)
                assert alignment.shape[1] >= time_length
                self.alignment = alignment
            elif isinstance(self.alignment, Alignment):
                if len(self.alignment.items) + 1 != len(plot.output):  # except EOS
                    return False
                assert self.time_reduction == 1 and self.add_extra_time == 0 and self.add_extra_time_red == 0
                alignment = numpy.zeros((len(plot.output), time_length), dtype="float32")
                for i, item in enumerate(self.alignment.items):
                    alignment[i, item.start_frame : item.end_frame + 1] = 1.0
                self.alignment = alignment
            else:
                raise TypeError("%s: unexpected alignment %r, type %s" % (self, self.alignment, type(self.alignment)))
            assert isinstance(self.alignment, numpy.ndarray)
            return True

        def set_mapped_alignment(self):
            """
            :return: whether to use this
            :rtype: bool
            """
            import seaborn
            from seaborn import light_palette

            if not self._check_alignment():
                return False
            if self.cmap is None:
                if self.cmap_palette is not None:
                    self.cmap = seaborn.color_palette()[self.cmap_palette]
                else:
                    plot = self.plot.plot_by_seq_tag[self.seq_tag]  # recent_plot might be different
                    self.cmap = plot._get_next_cmap()
            cmap = light_palette(self.cmap, as_cmap=True, n_colors=50)
            self.mapped_alignment = cmap(self.alignment, alpha=1)  # type: numpy.ndarray
            return True

    class Plot:
        def __init__(self, seq_tag, parent):
            """
            :param str seq_tag:
            :param AlignmentPlotter parent:
            """
            self.parent = parent
            self.seq_tag = seq_tag
            self.model_names = []  # type: typing.List[str]
            self.use_tex = False
            self.time_length = None  # type: typing.Optional[int]
            self.output = None  # type: typing.Optional[typing.List[str]]  # e.g. BPE units by the decoder
            self.alignments = {}  # type: typing.Dict[str,AlignmentPlotter.Alignment]
            self.time_alignment = None  # type: typing.Optional[Alignment]
            self._palette_idx = -1
            self.style_type = "paper"
            self.style_figsize = (8, 5)
            self.style_titlesize = None

        def set_time_alignment(self, time_alignment):
            """
            :param Alignment time_alignment:
            :return: whether this was added
            :rtype: bool
            """
            if self.time_alignment:
                assert self.time_alignment == time_alignment
                return False
            self.time_alignment = time_alignment
            AlignmentPlotter.Alignment(
                plot=self.parent, key="00_00_hybrid", human="Hybrid", alignment=time_alignment, cmap_palette=2
            ).add_to_plot()
            return True

        def add_model_name(self, name):
            """
            :param str name:
            :return: key prefix if new, else None
            :rtype: str|None
            """
            if name in self.model_names:
                return None
            self.model_names.append(name)
            return "%02i" % len(self.model_names)

        def _get_next_cmap(self):
            import seaborn

            self._palette_idx += 1
            return [seaborn.color_palette()[i] for i in [0, 5]][self._palette_idx]

        def _set_style(self):
            """
            Sets the matplotlib style according to the parameters.
            """
            style = self.style_type
            figsize = self.style_figsize
            titlesize = self.style_titlesize
            import matplotlib.pyplot as plt

            assert style in ("talk", "paper")
            import warnings

            # matplotlib warns about the font family,
            # but the rendering is done by latex.
            warnings.filterwarnings(
                "ignore", category=UserWarning, module="matplotlib", message=r"findfont: Font family.*"
            )
            # seaborn-paper: axes.titlesize: 9.6
            # seaborn-talk: axes.titlesize: 15.6
            if titlesize is None:
                if style == "paper":
                    titlesize = 10
                elif style == "talk":
                    titlesize = 16
            legendsize = titlesize * 0.6
            labelsize = titlesize * 0.8

            # noinspection PyTypeChecker
            plt.style.use(
                [
                    "seaborn-%s" % style,
                    {
                        "text.usetex": self.use_tex,
                        "text.latex.preamble": [r"\usepackage{times}"],
                        "font.family": "arial",
                        "font.size": labelsize,
                        "legend.facecolor": "white",
                        "figure.dpi": 120,
                        "figure.figsize": figsize,
                        "figure.autolayout": True,
                        "axes.titlesize": titlesize,
                        "axes.labelsize": labelsize,
                        "xtick.labelsize": labelsize,
                        "ytick.labelsize": labelsize,
                        "legend.fontsize": legendsize,
                    },
                ]
            )

        def _string_latex_compatible(self, output):
            """
            :param str output:
            :rtype: str
            """
            if not self.use_tex:
                return output
            return output.replace("<", r"\textless").replace(">", r"\textgreater").replace("_", r"\_")

        def _list_latex_compatible(self, output):
            """
            :param list[str] output:
            :rtype: list[str]
            """
            assert isinstance(output, list)
            return [self._string_latex_compatible(o) for o in output]

        def _add_plot_seq_title(self, ax, prepend="", **title_kws):
            """
            Adds a sequence title to the plot

            :param matplotlib.pyplot.Axes ax:
            :param str prepend:
            """
            from textwrap import wrap

            output = " ".join(self.output)
            output = output.replace("@@ ", "")
            output = output.replace("</s>", "")
            output = output.replace("<s>", "")
            title = '%s%s:\n"%s"' % (prepend, self.seq_tag, "\n".join(wrap(output, 60)))
            title = self._string_latex_compatible(title)
            ax.set_title(title, **title_kws)

        def _add_plot_time_title(self, ax, rotate_45=False, **font_kwargs):
            """
            Show phonemes or words above the plot, aligned to their time position.

            :param matplotlib.pyplot.Axes ax:
            :param bool rotate_45:
            """
            start_y = -1.0
            assert self.time_alignment
            for i, item in enumerate(self.time_alignment.items):
                phoneme = item.value
                phoneme = phoneme.lower().replace("[silence]", "sil")
                x1 = item.start_frame
                x2 = item.end_frame + 1.0
                p = start_y

                t = ax.text(x=x1, y=p, s=phoneme, **font_kwargs)
                if rotate_45:
                    t.set_rotation(45)

                ax.axvline(x=x1 - 0.5, color="gray", linestyle=":")
                ax.axvline(x=x2 - 0.5, color="gray", linestyle=":")
            ax.axvline(x=self.time_length - 0.5, color="black", linestyle="-")

        def _add_plot_decoder_output(self, ax, hide_prev_label=True, stride=1, draw_lines=True):
            """
            Shows the decoder output on the left axis.

            :param matplotlib.pyplot.Axes ax:
            :param bool hide_prev_label:
            :param int stride:
            :param bool draw_lines:
            """
            from matplotlib.ticker import AutoMinorLocator, NullFormatter, FixedFormatter

            if hide_prev_label:
                ax.set_ylabel("")
            assert self.output is not None
            output = self._list_latex_compatible(self.output)
            ax.set_yticklabels(output, verticalalignment="bottom")

            ticks = numpy.arange(len(output) * stride + stride)[::stride] - 0.5
            ax.set_yticks(ticks)

            # tick labels, we have to adjust so that the label is centered
            # minor ticks are in-between major ticks
            ax.yaxis.set_major_formatter(NullFormatter())
            ax.yaxis.set_minor_locator(AutoMinorLocator(2))
            ax.yaxis.set_minor_formatter(FixedFormatter(output))
            ax.yaxis.set_ticks_position("none")  # hide labels only

            # draw output boundaries (horizontal lines)
            if draw_lines:
                for i, output_bpe in enumerate(output):
                    ax.axhline(y=i * stride + stride - 0.5, color="gray", linestyle="-")

        def do_plot(self, save_filename=None, show=None, with_title=True):
            """
            Plot.

            :param str|None save_filename:
            :param bool|None show:
            :param bool with_title:
            """
            import matplotlib.pyplot as plt
            from seaborn import light_palette
            from mpl_toolkits.axes_grid1 import make_axes_locatable
            import matplotlib.patches as mpatches
            from matplotlib.cm import ScalarMappable
            import matplotlib.patches as patches

            for align_type, d in sorted(self.alignments.items()):
                assert isinstance(d, AlignmentPlotter.Alignment)
                if not d.set_mapped_alignment():
                    self.alignments.pop(align_type)

            stride = len(self.alignments)
            self._set_style()

            fig, ax = plt.subplots()
            assert isinstance(ax, plt.Axes)

            # mix alignment matrices (interleave)
            # we have matrices of form [dec_len, enc_len]
            # for 4x alignments, we get [4*dec_len, enc_len]
            multi_dec = stride * len(self.output)

            assert self.time_length is not None
            max_time_len = max([d.mapped_alignment.shape[1] for _, d in self.alignments.items()])
            alignment_matrix = numpy.ones(shape=(multi_dec, max_time_len, 4))  # RGBA
            offset = 0
            for align_type, d in sorted(self.alignments.items()):
                assert isinstance(d, AlignmentPlotter.Alignment)
                alignment_matrix[offset::stride, : d.mapped_alignment.shape[1]] = d.mapped_alignment
                offset += 1

            # fixes the issue where the background is slightly colored
            # the value for clipping is related to the cmap levels
            alignment_matrix[alignment_matrix > 0.9] = 1

            ax.imshow(alignment_matrix, aspect="auto")

            ax.xaxis.tick_bottom()
            ax.set_xlabel("Input sequence")
            ax.set_ylabel("Output sequence")

            divider = make_axes_locatable(ax)
            legend_patches = []
            # legend_patches.append(dotted_line)
            prev_colorbar_palettes = set()  # type: typing.Set[int]
            idx = 0
            for align_type, d in sorted(self.alignments.items()):
                assert isinstance(d, AlignmentPlotter.Alignment)
                patch = mpatches.Patch(color=d.cmap, label=d.human)
                legend_patches.append(patch)
                if not d.colorbar:
                    continue
                if d.cmap_palette is not None:
                    if d.cmap_palette in prev_colorbar_palettes:
                        continue
                    prev_colorbar_palettes.add(d.cmap_palette)
                cax = divider.append_axes("right", size="5%", pad=0.1 if idx == 0 else 0)
                cmap = light_palette(d.cmap, as_cmap=True)
                mappable = ScalarMappable(cmap=cmap)
                mappable.set_array(d.mapped_alignment)
                ticks = [0.0, 0.2, 0.4, 0.6, 0.8, 1]
                plt.colorbar(mappable, cax=cax, ticks=ticks)
                idx += 1

            ax.legend(handles=legend_patches, frameon=True, handlelength=3, loc="best")

            self._add_plot_decoder_output(ax, stride=stride, hide_prev_label=True)
            self._add_plot_time_title(ax)
            if with_title:
                self._add_plot_seq_title(ax, y=1.15)

            offset = 0
            for align_type, d in sorted(self.alignments.items()):
                assert isinstance(d, AlignmentPlotter.Alignment)
                for i in range(len(self.output)):
                    arg_max = numpy.argmax(d.alignment[i])
                    if d.alignment[i, arg_max] > 0:
                        window_left, window_right = None, None  # inclusive,exclusive
                        if d.alignment[i, 0] == 0.0:  # exact 0.0 -> likely there is a window
                            window_left = numpy.where(numpy.cumsum(d.alignment[i, :arg_max]) == 0.0)[0][-1] + 1
                            assert window_left <= arg_max and d.alignment[i, window_left - 1] == 0.0
                            assert d.alignment[i, window_left] > 0 and all(d.alignment[i, :window_left] == 0.0)
                        if d.alignment[i, -1] == 0.0:  # exact 0.0 -> likely there is a window
                            window_right = (
                                numpy.where(numpy.cumsum(d.alignment[i, arg_max + 1 :][::-1])[::-1] == 0.0)[0][0]
                                + arg_max
                                + 1
                            )
                            assert window_right > arg_max and d.alignment[i, window_right] == 0.0
                            assert d.alignment[i, window_right - 1] > 0 and all(d.alignment[i, window_right:] == 0.0)
                        if window_left is not None or window_right is not None:
                            if window_left is None:
                                window_left = 0
                            if window_right is None:
                                window_right = d.alignment.shape[1]
                            rect = patches.Rectangle(
                                (window_left - 0.5, offset + i * stride - 0.5 - 0.05),
                                window_right - window_left - 0.05,
                                1.0 - 0.05,
                                linewidth=0.5,
                                edgecolor="lightgray",
                                facecolor="none",
                            )
                            ax.add_patch(rect)
                offset += 1

            if save_filename:
                plt.savefig(save_filename)
                print("Saved to file:", save_filename)
            if show is None:
                show = not save_filename
            if show:
                plt.show(block=True)


def read_hdf(
    hdf_filename,
    rec_layer_name="output",
    lexicon=None,
    apply_bpe_merge=False,
    remove_sentence_end=None,
    phone_alignment_file_pattern=None,
    pos_choice=None,
    pos_time_reduction_factor=1,
    add_extra_time_frames=0,
    add_extra_time_red_frames=0,
    plot_replace_names=None,
    plotter=None,
    plot_name_prefix="",
):
    """
    :param str hdf_filename:
    :param str rec_layer_name:
    :param str|None lexicon:
    :param bool apply_bpe_merge:
    :param str|None remove_sentence_end:
    :param str|None phone_alignment_file_pattern:
    :param str|None pos_choice:
    :param int pos_time_reduction_factor:
    :param int add_extra_time_frames:
    :param int add_extra_time_red_frames:
    :param bool|None|AlignmentPlotter plotter:
    :param str plot_name_prefix:
    :param dict[str,str]|None plot_replace_names:
    :return: plotter
    :rtype: AlignmentPlotter|None
    """
    assert os.path.exists(hdf_filename), "hdf file not dumped?"

    if lexicon:
        lexicon = Lexicon(filename=lexicon)

    phone_alignments = None  # name -> ...
    if phone_alignment_file_pattern:
        phone_alignments = parse_phone_alignments(phone_alignment_file_pattern)

    f = h5py.File(hdf_filename, "r")
    num_steps = f["seqLengths"].shape[0]  # we used dump_whole_batches
    print("Num steps:", num_steps)
    keys = sorted(f["targets/size"].attrs.keys())  # type: typing.List[str]
    keys.insert(0, "data")
    print("Data keys:")
    pprint(keys)
    assert len(keys) == f["seqLengths"].shape[1]  # seq-lens for each data-key and 'data'
    offsets = {key: 0 for key in keys}
    key_prefix = "%s__" % rec_layer_name
    assert all([key.startswith(key_prefix) for key in keys if key not in ["data", "seq_sizes", "seq_tags"]])
    labels = f["labels"]
    print("Labels dim:", labels.shape[0])
    if not labels.shape[0]:
        labels = None

    if plotter:
        if isinstance(plotter, bool):
            plotter = AlignmentPlotter()
        assert isinstance(plotter, AlignmentPlotter)
    if plot_replace_names is None:
        plot_replace_names = {}

    for step in range(num_steps):
        print("Step:", step)

        raw_values = {}
        for key_idx, key in enumerate(keys):
            raw_num_frames = f["seqLengths"][step, key_idx]
            offset = offsets[key]
            offsets[key] += raw_num_frames
            if key == "data":
                group = f["inputs"]
            else:
                group = f["targets/data/%s" % key]
            raw_value = group[offset:raw_num_frames]
            raw_values[key] = raw_value

        seq_sizes = raw_values["seq_sizes"]
        seq_tags = raw_values["seq_tags"]
        num_seqs = seq_sizes.shape[0]
        assert seq_tags.shape == (num_seqs,) == seq_sizes.shape[:1]
        print("  Num seqs:", num_seqs)

        choice_layer_names = {key[: -len("_src_beams_raw")] for key in keys if key.endswith("_src_beams_raw")}
        assert all([name.startswith(key_prefix) for name in choice_layer_names])
        choice_layer_names = {name[len(key_prefix) :] for name in choice_layer_names}
        choice_layer_names = sorted(choice_layer_names)
        if "output" in choice_layer_names:
            # Add it at the end. This is an implicit assumption about the choice order.
            choice_layer_names.remove("output")
            choice_layer_names.append("output")
        print("  Choice layers:", choice_layer_names)
        if pos_choice:
            assert pos_choice in choice_layer_names
        choices_num_seqs = {
            name: raw_values["%s%s_src_beams_raw_seq_lens" % (key_prefix, name)].shape[0] for name in choice_layer_names
        }
        assert all([n % num_seqs == 0 for (_, n) in choices_num_seqs.items()])
        choices_beam_sizes = {name: n // num_seqs for (name, n) in choices_num_seqs.items()}
        print("  Choice beam sizes:", choices_beam_sizes)
        output_dim = f["targets/size"].attrs["%s%s_value_final" % (key_prefix, choice_layer_names[-1])][0]
        print("  %r dim:" % choice_layer_names[-1], output_dim)
        if labels is not None:
            assert labels.shape[0] == output_dim

        for seq_idx, seq_tag in enumerate(seq_tags):
            print("  Seq idx:", seq_idx)
            print("    Seq tag:", seq_tag)

            last_beam_size = choices_beam_sizes[choice_layer_names[-1]]
            last_choice_seq_lens = raw_values["%s%s_beam_scores_raw_seq_lens" % (key_prefix, choice_layer_names[-1])]
            assert last_choice_seq_lens.shape == (num_seqs * last_beam_size, 1)
            last_choice_seq_lens = numpy.reshape(last_choice_seq_lens, (num_seqs, last_beam_size))[
                seq_idx
            ]  # (beam_size,)
            print("    N-best seq lens:", list(last_choice_seq_lens))
            max_seq_len = max(last_choice_seq_lens)

            if plotter:
                plotter.set_recent_seq_tag(seq_tag)

            for i in range(max_seq_len):
                print("    Frame i:", i)

                for choice_name in choice_layer_names:
                    beam_size = choices_beam_sizes[choice_name]
                    print("      Choice:", choice_name, ", beam size:", beam_size)
                    choice_seq_lens = raw_values["%s%s_beam_scores_raw_seq_lens" % (key_prefix, choice_name)]
                    assert choice_seq_lens.shape == (num_seqs * beam_size, 1)
                    choice_max_seq_len = max(choice_seq_lens.flat)

                    beam_scores = raw_values["%s%s_beam_scores_raw" % (key_prefix, choice_name)]
                    assert beam_scores.shape == (num_seqs * beam_size * choice_max_seq_len,)
                    beam_scores = numpy.reshape(beam_scores, (num_seqs, beam_size, choice_max_seq_len))[seq_idx, :, i]
                    print("        Beam scores:", list(beam_scores))

                    src_beams = raw_values["%s%s_src_beams_raw" % (key_prefix, choice_name)]
                    assert src_beams.shape == (num_seqs * beam_size * choice_max_seq_len,)
                    src_beams = numpy.reshape(src_beams, (num_seqs, beam_size, choice_max_seq_len))[seq_idx, :, i]
                    print("        Src beam indices:", list(src_beams))

                    choice_values = raw_values["%s%s_value_raw" % (key_prefix, choice_name)]
                    assert choice_values.shape == (num_seqs * beam_size * choice_max_seq_len,)
                    choice_values = numpy.reshape(choice_values, (num_seqs, beam_size, choice_max_seq_len))[
                        seq_idx, :, i
                    ]
                    print("        Choice values:", list(choice_values))

            plot_key_prefix = None
            if plotter:
                plot_key_prefix = plotter.recent_plot.add_model_name(plot_name_prefix)
                assert plot_key_prefix

            for n in range(last_beam_size):
                final_beam_scores = raw_values["%sfinal_beam_scores" % key_prefix]
                assert final_beam_scores.shape == (num_seqs * last_beam_size,)
                final_beam_scores = numpy.reshape(final_beam_scores, (num_seqs, last_beam_size))
                print("    Final hyp %i (out of %i), score %f:" % (n, last_beam_size, final_beam_scores[seq_idx, n]))
                for choice_name in choice_layer_names:
                    choice_seq_lens = raw_values["%s%s_beam_scores_final_seq_lens" % (key_prefix, choice_name)]
                    assert choice_seq_lens.shape == (num_seqs * last_beam_size, 1)
                    choice_max_seq_len = max(choice_seq_lens.flat)

                    choice_values = raw_values["%s%s_value_final" % (key_prefix, choice_name)]
                    assert choice_values.shape == (num_seqs * last_beam_size * choice_max_seq_len,)
                    choice_values = numpy.reshape(choice_values, (num_seqs, last_beam_size, choice_max_seq_len))
                    choice_values = list(choice_values[seq_idx, n, : last_choice_seq_lens[n]])
                    print("      %s:" % choice_name, choice_values)

                    if choice_name == pos_choice:
                        hard_att_align = [v * pos_time_reduction_factor for v in choice_values]
                        print("      %s pos with factor %i:" % (choice_name, pos_time_reduction_factor), hard_att_align)
                        if plotter and plot_key_prefix:
                            AlignmentPlotter.Alignment(
                                plot=plotter,
                                key="%s_02_hard_att_%s" % (plot_key_prefix, choice_name),
                                human="%s%s" % (plot_name_prefix, plot_replace_names.get(choice_name, choice_name)),
                                cmap_palette=4,
                                alignment=choice_values,
                                time_reduction=pos_time_reduction_factor,
                                add_extra_time=add_extra_time_frames,
                                add_extra_time_red=add_extra_time_red_frames,
                            ).add_to_plot(overwrite=True)

                    if choice_name == choice_layer_names[-1] and labels is not None:
                        label_str = " ".join([labels[l] for l in choice_values])
                        print("      %s labels:" % choice_name, label_str)
                        if plotter:
                            plotter.recent_plot.output = label_str.split()
                        if remove_sentence_end and label_str.endswith(" %s" % remove_sentence_end):
                            label_str = label_str[: -len(remove_sentence_end) - 1]
                        if apply_bpe_merge:
                            label_str = label_str.replace("@@ ", "")
                            print("      %s labels (after BPE merge):" % choice_name, label_str)
                        if lexicon:
                            words = label_str.split()
                            phone_alignment = select_phone_alignment(phone_alignments=phone_alignments, seq_tag=seq_tag)
                            if plotter:
                                plotter.recent_plot.time_length = phone_alignment.items[-1].end_frame + 1
                            word_alignment = phone_alignment_to_word_alignment(
                                lexicon=lexicon, words=words, phone_alignment=phone_alignment
                            )
                            print("      word alignment:", word_alignment)
                            # We found word alignment -> this output seq is ground truth.
                            if plotter and word_alignment and plot_key_prefix:
                                plotter.recent_plot.set_time_alignment(word_alignment)
                                spatial_sm_value_keys = {
                                    key[: -len("_spatial_sm_value_final")]
                                    for key in keys
                                    if key.endswith("_spatial_sm_value_final")
                                }
                                assert all([name.startswith(key_prefix) for name in spatial_sm_value_keys])
                                spatial_sm_value_keys = {name[len(key_prefix) :] for name in spatial_sm_value_keys}
                                print("      Spatial softmax keys:", spatial_sm_value_keys)
                                for key in spatial_sm_value_keys:
                                    spatial_seq_lens = raw_values[
                                        "%s%s_spatial_sm_value_final_seq_lens" % (key_prefix, key)
                                    ]
                                    assert spatial_seq_lens.shape[0] == num_seqs * last_beam_size
                                    assert spatial_seq_lens.ndim == 2
                                    assert spatial_seq_lens.shape[1] in {
                                        2,
                                        3,
                                    }  # (dec-len,enc-len) tuples, or (dec-len,enc-len,head)
                                    max_dec_len = max(spatial_seq_lens[:, 0])
                                    max_enc_len = max(spatial_seq_lens[:, 1])
                                    spatial_seq_lens = numpy.reshape(
                                        spatial_seq_lens, (num_seqs, last_beam_size, spatial_seq_lens.shape[-1])
                                    )
                                    if spatial_seq_lens.shape[-1] == 2:
                                        dec_len, enc_len = spatial_seq_lens[seq_idx, n]
                                        num_heads = 1
                                    else:
                                        dec_len, enc_len, num_heads = spatial_seq_lens[seq_idx, n]
                                    spatial_sm_value = raw_values["%s%s_spatial_sm_value_final" % (key_prefix, key)]
                                    assert spatial_sm_value.shape == (
                                        num_seqs * last_beam_size * max_dec_len * max_enc_len * num_heads,
                                    )
                                    spatial_sm_value = numpy.reshape(
                                        spatial_sm_value,
                                        (num_seqs, last_beam_size, max_dec_len, max_enc_len, num_heads),
                                    )
                                    spatial_sm_value = spatial_sm_value[seq_idx, n, :dec_len, :enc_len]
                                    if spatial_seq_lens.shape[-1] == 2 or num_heads == 1:
                                        AlignmentPlotter.Alignment(
                                            plot=plotter,
                                            key="%s_01_sm_%s" % (plot_key_prefix, key),
                                            human="%s%s" % (plot_name_prefix, plot_replace_names.get(key, key)),
                                            alignment=spatial_sm_value[:, :, 0],
                                            time_reduction=pos_time_reduction_factor,
                                            add_extra_time=add_extra_time_frames,
                                            add_extra_time_red=add_extra_time_red_frames,
                                            cmap_palette=0,
                                            colorbar=True,
                                        ).add_to_plot()
                                    else:
                                        for h in range(num_heads):
                                            AlignmentPlotter.Alignment(
                                                plot=plotter,
                                                key="%s_01_sm_%s_%02i" % (plot_key_prefix, key, h),
                                                human="%s%s, head %i"
                                                % (plot_name_prefix, plot_replace_names.get(key, key), h),
                                                alignment=spatial_sm_value[:, :, h],
                                                time_reduction=pos_time_reduction_factor,
                                                add_extra_time=add_extra_time_frames,
                                                add_extra_time_red=add_extra_time_red_frames,
                                                cmap_palette=0,
                                                colorbar=True,
                                            ).add_to_plot()
                                plot_key_prefix = None

    return plotter


def main():
    arg_parser = argparse.ArgumentParser(description="Visualize search scores")
    arg_parser.add_argument("hdf", help="dumped via debug-dump-search-scores.py")
    arg_parser.add_argument("--rec_layer_name", default="output")
    arg_parser.add_argument("--lexicon", help="xml file")
    arg_parser.add_argument("--apply_bpe_merge", action="store_true", help="replace '@@ ' by ''")
    arg_parser.add_argument("--remove_sentence_end", help="e.g. '<s>'")
    arg_parser.add_argument("--phone_alignment_file_pattern", help="e.g. 'align-*.txt', stdout from RASR archiver")
    arg_parser.add_argument("--pos_choice", help="e.g. 't' in decoder")
    arg_parser.add_argument("--pos_time_reduction_factor", type=int, default=1)
    arg_parser.add_argument("--add_extra_time_frames", type=int, default=0)
    arg_parser.add_argument("--plot", action="store_true")
    args = arg_parser.parse_args()

    plotter = read_hdf(
        hdf_filename=args.hdf,
        rec_layer_name=args.rec_layer_name,
        lexicon=args.lexicon,
        apply_bpe_merge=args.apply_bpe_merge,
        remove_sentence_end=args.remove_sentence_end,
        phone_alignment_file_pattern=args.phone_alignment_file_pattern,
        pos_choice=args.pos_choice,
        pos_time_reduction_factor=args.pos_time_reduction_factor,
        add_extra_time_frames=args.add_extra_time_frames,
        plotter=args.plot,
    )

    if plotter:
        for seq_tag, plot in sorted(plotter.plot_by_seq_tag.items()):
            assert isinstance(plot, AlignmentPlotter.Plot)
            plot.do_plot()


if __name__ == "__main__":
    try:
        from returnn.util import better_exchook

        better_exchook.install()
    except ImportError:
        better_exchook = None
    main()



"""
Tensor utils.
"""

from __future__ import annotations
import numpy
import torch
from returnn.tensor import Tensor


def tensor_numpy_to_torch_(x: Tensor[numpy.ndarray]):
    """
    torch.from_numpy() on Tensor, including dims
    """
    if x.raw_tensor is None or isinstance(x.raw_tensor, torch.Tensor):
        pass
    else:
        assert isinstance(x.raw_tensor, numpy.ndarray)
        x.raw_tensor = torch.from_numpy(x.raw_tensor)
    for dim in x.dims:
        if dim.dyn_size_ext:
            tensor_numpy_to_torch_(dim.dyn_size_ext)


def tensor_torch_to_numpy_(x: Tensor[torch.Tensor]):
    """
    .numpy() on Tensor, including dims
    """
    if x.raw_tensor is None or isinstance(x.raw_tensor, numpy.ndarray):
        pass
    else:
        assert isinstance(x.raw_tensor, torch.Tensor)
        x.raw_tensor = x.raw_tensor.detach().cpu().numpy()
    for dim in x.dims:
        if dim.dyn_size_ext:
            tensor_torch_to_numpy_(dim.dyn_size_ext)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_neg():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return -x

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)



"""
Random number generator utilities.

Note on the global seed:

In the TF engine, we have this in ``_init_network``:

.. code-block:: python

    tf_random_seed = 42
    net_random_seed = epoch
    if self.config.opt_typed_value("random_seed", None):
        seed = self.config.int("random_seed", None)
        net_random_seed = (epoch * 3 + seed * 5 + 7) % (2**31)
        tf_random_seed = (net_random_seed * 2 + 3) % (2**31)
    tf_compat.v1.set_random_seed(tf_random_seed)

``_init_network`` is only called in the beginning
and then only when needed (e.g. different model due to pretrain).

The ``net_random_seed`` is used inside the :class:`TFNetwork` for:

.. code-block:: python

    self.random = numpy.random.RandomState(rnd_seed)

This in turns is used for param init in the TF layers, like:

.. code-block:: python

    fwd_weights_initializer = get_initializer(
        forward_weights_init, seed=self.network.random.randint(2**31), eval_local_ns={"layer": self}
    )

It is also used by the :class:`RandomLayer` when ``static=True``:

.. code-block:: python

    # static is True
    if seed is None:
        seed = self.network.random.randint(2**31, size=[32], dtype="uint32")

So, it means, with ``static=True``, the random numbers are always the same in each execution step.
To reflect that in an eager-based backend,
we need to reset the static-random-state in the beginning of each step.

"""

from __future__ import annotations
from typing import Optional, Union, Sequence, Dict
import numpy
from returnn.tensor import Tensor, Dim
from ._backend import global_backend as _global_backend


__all__ = [
    "set_random_seed",
    "get_random_state",
    "set_random_state",
    "reset_step_random_state",
    "get_static_step_based_seed",
    "random",
    "random_uniform",
]


def set_random_seed(seed: int):
    """
    Call this at the beginning of the program
    (after the RF backend was selected),
    or when the model and computation graph is supposed to be reinitialized.

    This initializes the random state of the backend and also the step-based random state.

    This is *not* expected to be called after each epoch or step.

    :param seed: should depend on epoch or step
    """
    _global_backend.set_random_seed(seed)
    global _step_rnd_seed
    _step_rnd_seed = (seed * 5393 + 1187) % (2**31)
    reset_step_random_state()


def get_random_state() -> Dict[str, bytes]:
    """
    :return: current random state, for serialization, to be able to restore it later
    """
    return _global_backend.get_random_state()


def set_random_state(state: Dict[str, bytes]):
    """
    Recovers the random state.

    There are many potential cases where we cannot recover the state
    (e.g. different backend version, different hardware, ...),
    In this case, a run without interruption is not the same as a run with interruption.

    We still assume that :func:`set_random_seed` was called before in any case.

    :param state: as returned by :func:`get_random_state`
    """
    # Always call this first.
    _global_backend.set_random_state(state)


_step_rnd_seed = 42
_step_rnd = numpy.random.RandomState(_step_rnd_seed)


def reset_step_random_state():
    """
    When ``static=True`` is used in :func:`random`,
    the random state is reset to the beginning of the step.
    So this should be called in the beginning of each step.
    Also see the module docstring.
    """
    _step_rnd.seed(_step_rnd_seed)


def get_static_step_based_seed(*, size=None) -> Union[int, numpy.ndarray]:
    """
    :return: from the static step-based random state, get a seed
    """
    return _step_rnd.randint(2**31, size=size)


def random(
    *,
    dims: Sequence[Dim],
    dtype: Optional[str] = None,
    sparse_dim: Optional[Dim] = None,
    distribution: str,
    mean: Optional[Union[int, float, Tensor]] = None,
    stddev: Optional[Union[int, float, Tensor]] = None,
    bound: Optional[Union[int, float, Tensor]] = None,
    minval: Optional[Union[int, float, Tensor]] = None,
    maxval: Optional[Union[int, float, Tensor]] = None,
    seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = None,
    algorithm: Optional[str] = None,
    explicit_state: Optional[Tensor] = None,
    auto_update_state: Optional[bool] = None,
    static: Optional[bool] = None,
    out: Optional[Tensor] = None,
) -> Tensor:
    """
    Generates random numbers from uniform or normal or truncated normal distribution.

    There will be no gradients to mean, stddev, bound, minval, maxval!

    In case of TensorFlow:
    This uses the TensorFlow stateless random ops internally, i.e. all the state handling is explicit.
    The state var can be explicitly provided and initialized via :class:`RandomStateInitLayer`,
    or when not provided it will be automatically created.

    There are two possible distinct use cases:

    - For any randomness in the model, e.g. dropout. So each ``session.run`` step will produce a new random number
      and advance the random state.
    - To initialize parameters via the config, using :class:`VariableLayer` with the ``init_by_layer`` option.
      This will only be called once when initializing the parameters.
      For this use case, we do not want to keep a random state var.
      You can just pass ``static=False``.
      Alternatively you could also pass the output of a :class:`RandomStateInitLayer` as ``state``.

    :param Sequence[Dim] dims:
    :param str dtype:
    :param sparse_dim:
    :param str distribution: "uniform", "normal" or "truncated_normal"
    :param int|float|Tensor|None mean:
    :param int|float|Tensor|None stddev:
    :param int|float|Tensor|None bound: for uniform, defining the range [-bound, bound)
    :param int|float|Tensor|None minval: for uniform
    :param int|float|Tensor|None maxval: for uniform
    :param int|list[int]|numpy.ndarray|None seed: If not given, uses self.network.random.randint,
      i.e. then it is controlled by the global seed setting, and every layer would get its own seed.
      If you specify it explicitly, make sure every :class:`RandomLayer` uses a different seed,
      otherwise you would get the same random numbers everywhere.
    :param str|tf.random.Algorithm|None algorithm: see :class:`RandomStateInitLayer`
    :param Tensor|None explicit_state: You can pass the state explicitly here.
      If not given, will be created automatically, and updated automatically.
      You could pass a :class:`VariableLayer` with initial value via :class:`RandomStateInitLayer`,
      or directly a :class:`RandomStateInitLayer`.
      If auto_update_state is True, it must be a variable,
      and every time a new random number is created, this variable is updated.
      Otherwise (default), it will not be updated automatically.
    :param bool|None auto_update_state: only used when you pass an explicit state
    :param bool|None static: if no state at all should be used. it just relies on the seed then.
    :param out: if given, will directly write into it, if possible by backend
    :return: random values
    """
    if explicit_state is None:
        if static is None:
            static = False
        assert isinstance(static, bool)
        if static:
            if seed is None:
                seed = get_static_step_based_seed()
    return _global_backend.random(
        dims=dims,
        dtype=dtype,
        sparse_dim=sparse_dim,
        distribution=distribution,
        mean=mean,
        stddev=stddev,
        bound=bound,
        minval=minval,
        maxval=maxval,
        seed=seed,
        algorithm=algorithm,
        explicit_state=explicit_state,
        auto_update_state=auto_update_state,
        static=static,
        out=out,
    )


def random_uniform(
    *,
    dims: Sequence[Dim],
    dtype: Optional[str] = None,
    sparse_dim: Optional[Dim] = None,
    minval: Union[int, float, Tensor],
    maxval: Union[int, float, Tensor],
    seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = None,
    algorithm: Optional[str] = None,
    explicit_state: Optional[Tensor] = None,
    auto_update_state: Optional[bool] = None,
    static: Optional[bool] = None,
    out: Optional[Tensor] = None,
):
    """
    See :func:`random`. :func:`random` with ``distribution="uniform"``.
    """
    return random(
        dims=dims,
        dtype=dtype,
        sparse_dim=sparse_dim,
        distribution="uniform",
        minval=minval,
        maxval=maxval,
        seed=seed,
        algorithm=algorithm,
        explicit_state=explicit_state,
        auto_update_state=auto_update_state,
        static=static,
        out=out,
    )



"""
High-level backend for RETURNN layers
"""

from __future__ import annotations
from typing import Union, Sequence, Optional, Any, Tuple, Dict
import contextlib
import numpy
import tensorflow as tf
import returnn.tf.compat as tf_compat

from returnn.util.basic import NotSpecified
from returnn.tensor import Tensor, Dim

# noinspection PyProtectedMember
from returnn.frontend._backend import Backend

from .layer import Layer
from .. import frontend_layers as rfl
from ... import frontend as rf
from . import dims as _dims
from returnn.frontend import RawTensorTypes


# Ignore this warning until we really expect that we implemented everything.
# noinspection PyAbstractClass
class ReturnnLayersBackend(Backend[Layer]):
    """
    RETURNN layers backend (using TF), where raw_tensor represents a RETURNN layer
    """

    RawTensorType = Layer
    is_tensorflow = True

    @staticmethod
    def executing_eagerly() -> bool:
        """executing eagerly"""
        return False

    @staticmethod
    def set_random_seed(seed: int):
        """
        :param seed:
        """
        tf_compat.v1.set_random_seed(seed)

    @staticmethod
    def get_random_state() -> Dict[str, bytes]:
        """
        :return: random state
        """
        # Not sure... we could cover all state vars used by RandomLayer...
        return {}

    @staticmethod
    def set_random_state(state: Dict[str, bytes]):
        """
        :param state: as returned by :func:`get_random_state`.
            This might not always be successful (e.g. different hardware, different backend version),
            so the calling code should always have called set_random_seed before to have the random generators
            in a reasonable fallback state.
        """
        assert not state  # not implemented... unexpected currently

    @staticmethod
    def get_dtype_name_raw(raw_tensor: Layer) -> str:
        """dtype"""
        return raw_tensor.tensor.dtype

    @staticmethod
    def as_dtype_raw(dtype_name: str) -> str:
        """dtype"""
        return dtype_name

    @staticmethod
    def get_ndim_raw(raw_tensor: Layer) -> int:
        """ndim"""
        return raw_tensor.tensor.batch_ndim

    @staticmethod
    def get_shape_raw(raw_tensor: Layer) -> Layer:
        """shape"""
        raise NotImplementedError

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: Layer) -> Tuple[Union[int, Layer]]:
        """shape"""
        raise NotImplementedError

    @staticmethod
    def get_known_shape_raw(raw_tensor: Layer) -> Tuple[Optional[int]]:
        """known shape"""
        return raw_tensor.tensor.batch_shape

    @staticmethod
    def set_known_shape_raw(raw_tensor: Layer, shape: Tuple[Optional[int]]) -> None:
        """set known shape"""
        pass  # ignore

    @staticmethod
    def fill_raw(shape: Union[Sequence[Union[int, Layer]], Layer], value: Union[Any, Layer]) -> Layer:
        """fill raw"""
        raise Exception("fill_raw not supported in layers backend because dim tags would be unknown")

    @staticmethod
    def compare_raw(a: Layer, kind: str, b: Layer) -> Layer:
        """compare"""
        raise Exception("compare_raw should not get called")

    @staticmethod
    def combine_raw(a: Layer, kind: str, b: Layer) -> Layer:
        """combine"""
        raise Exception("combine_raw should not get called")

    @staticmethod
    def reshape_raw(raw_tensor: Layer, shape: Union[Sequence[Union[int, Layer]], Layer]) -> Layer:
        """reshape_raw"""
        raise Exception("reshape_raw not supported in layers backend because dim tags would be unknown")

    @staticmethod
    def transpose(tensor: Tensor, perm: Sequence[Union[Dim, int]], *, allow_int: bool = False) -> Tensor:
        """transpose"""
        assert not allow_int  # not supported
        return rfl.make_layer({"class": "transpose", "from": tensor, "perm": perm}, name="transpose")

    @staticmethod
    def cast(tensor: Tensor, dtype: str) -> Tensor:
        """cast"""
        return rfl.make_layer({"class": "cast", "from": tensor, "dtype": dtype}, name="cast")

    @staticmethod
    def stop_gradient(tensor: Tensor) -> Tensor:
        """stop grad"""
        return rfl.make_layer({"class": "scaled_grad", "from": tensor, "scale": 0}, name="stop_gradient")

    @staticmethod
    def merge_dims(
        source: Tensor,
        *,
        dims: Sequence[Dim],
        out_dim: Optional[Dim] = None,
    ) -> Tuple[Tensor, Dim]:
        """
        Merges a list of axes into a single one. (Flatten the dims.)
        E.g. input is (batch, width, height, dim) and dims=(width,height), then we get (batch, width*height, dim).
        Or input is (batch, time, height, dim) and axes=(height,dim), then we get (batch, time, height*dim).

        :param source:
        :param dims:
        :param out_dim:
        :return: tensor, out_dim
        """
        if not isinstance(source, Tensor):
            raise TypeError(f"merge_dims: unexpected type for source {source!r}, need tensor")
        if out_dim is None:
            out_dim = dims[0]
            for d in dims[1:]:
                out_dim = out_dim * d
        layer = rfl.make_layer(
            {"class": "merge_dims", "from": source, "axes": dims, "out_dim": out_dim}, name="merge_dims"
        )
        return layer, out_dim

    @staticmethod
    def split_dims(
        source: Tensor,
        *,
        axis: Dim,
        dims: Sequence[Dim],
        pad_to_multiples: Optional[bool] = None,
        pad_value: Union[None, int, float] = None,
    ) -> Tensor:
        """split dims"""
        args = {}
        if pad_to_multiples is not None or pad_value is not None:
            args["pad_to_multiples"] = pad_to_multiples
            args["pad_value"] = pad_value
        args = {key: value for (key, value) in args.items() if value is not NotSpecified}
        return rfl.make_layer(
            {"class": "split_dims", "from": source, "axis": axis, "dims": dims, **args}, name="split_dims"
        )

    @staticmethod
    def reshape(source: Tensor, in_dims: Sequence[Dim], out_dims: Sequence[Dim]) -> Tensor:
        """reshape"""
        return rfl.make_layer(
            {
                "class": "reshape",
                "from": source,
                "in_dims": in_dims,
                "out_dims": out_dims,
                "extra_deps": rfl.get_dim_deps(out_dims),
            },
            name="reshape",
        )

    @staticmethod
    def split(source: Tensor, *, axis: Dim, out_dims: Sequence[Dim]) -> Tuple[Tensor, ...]:
        """split"""
        res = rfl.make_layer({"class": "split", "from": source, "axis": axis, "out_dims": out_dims}, name="split")

        src_axis_int = source.get_axis_from_description(axis)
        # noinspection PyProtectedMember
        return tuple(
            rfl._get_sub_layer(
                layer=res,
                name=str(i),
                data=source.copy_template_replace_dim_tag(
                    axis=src_axis_int, new_dim_tag=dim, name=f"{source.name}/split:{i}:{dim.description}"
                ),
            )
            for i, dim in enumerate(out_dims)
        )

    @staticmethod
    def expand_dim(source: Tensor, dim: Dim) -> Tensor:
        """expand dim"""
        # Some heuristic on the kind, which just determines where RETURNN puts the new axis.
        if source.have_feature_axis():
            axis = "spatial"
        elif dim.is_static():
            axis = "feature"
        else:
            axis = "spatial"
        return rfl.make_layer({"class": "expand_dims", "from": source, "axis": axis, "dim": dim}, name="expand_dims")

    @staticmethod
    def concat(
        *sources: Tuple[Tensor, Dim],
        allow_broadcast: bool = False,
        out_dim: Dim,
    ) -> Tensor:
        """concat"""
        opts = {}
        if allow_broadcast:
            opts["allow_broadcast"] = True
        out_dim = sum(d for _, d in sources)
        return rfl.make_layer(
            {"class": "concat", "from": sources, "out_dim": out_dim, **opts},
            name="concat",
        )

    @staticmethod
    def pad(
        source: Tensor,
        *,
        axes: Sequence[Dim],
        padding: Sequence[Tuple[Union[Dim, int], Union[Dim, int]]],
        out_dims: Sequence[Dim],
        mode: str = "constant",
        value: Union[rf.RawTensorTypes, Tensor] = None,
    ) -> Tensor:
        """pad"""
        assert isinstance(value, (int, float, type(None)))  # not implemented otherwise
        return rfl.make_layer(
            {
                "class": "pad",
                "from": source,
                "axes": axes,
                "padding": padding,
                "out_dims": out_dims,
                "mode": mode,
                "value": value,
            },
            name="pad",
        )

    @staticmethod
    def cum_concat_step(source: Tensor, *, prev_accum: Tensor, axis: Dim, out_spatial_dim: Dim) -> Tensor:
        """cum_concat_step"""
        return rfl.make_layer(
            {
                "class": "cum_concat",
                "from": source,
                "state": {"state": prev_accum},
                "out_spatial_dim": out_spatial_dim,
                "axis": axis,
            },
            name="cum_concat",
        )

    @staticmethod
    def activation(tensor: Tensor, func: str) -> Tensor:
        """activation"""
        return rfl.make_layer({"class": "activation", "activation": func, "from": tensor}, name=func)

    @staticmethod
    def activation_raw(raw_tensor: Layer, func: str) -> Layer:
        """activation"""
        return rfl.make_layer(
            {"class": "activation", "activation": func, "from": raw_tensor.tensor}, name=func
        ).raw_tensor

    @staticmethod
    def softmax(tensor: Tensor, *, axis: Dim, use_mask: bool = True) -> Tensor:
        """softmax"""
        args = {}
        if not use_mask:
            args["use_time_mask"] = False
        return rfl.make_layer({"class": "softmax_over_spatial", "axis": axis, "from": tensor, **args}, name="softmax")

    @staticmethod
    def log_softmax(tensor: Tensor, *, axis: Dim, use_mask: bool = True) -> Tensor:
        """log softmax"""
        args = {}
        if not use_mask:
            args["use_time_mask"] = False
        return rfl.make_layer(
            {"class": "softmax_over_spatial", "axis": axis, "from": tensor, "log_space": True, **args},
            name="log_softmax",
        )

    @staticmethod
    def softmax_cross_entropy_with_logits(*, logits: Tensor, targets: Tensor, axis: Dim):
        """
        Efficient cross entropy.

        :param logits: target estimates given as inputs to softmax (i.e. unnormalized)
        :param targets: probabilities, i.e. normalized, can also be sparse
        :param axis: class labels dim over which softmax is computed
        :return: cross entropy (same Dims as 'logits' but without 'axis')
        """
        if targets.sparse_dim:
            assert axis == targets.sparse_dim and axis in logits.dims
            return rfl.make_layer(
                {
                    "class": "sparse_softmax_cross_entropy_with_logits",
                    "logits": logits,
                    "targets": targets,
                    "axis": axis,
                },
                name="sparse_softmax_cross_entropy_with_logits",
            )
        else:  # dense targets
            assert axis in targets.dims and axis in logits.dims
            log_probs = rf.log_softmax(logits, axis=axis)
            return -rf.matmul(targets, log_probs, reduce=axis)

    @staticmethod
    def create_parameter_raw(tensor: rf.Parameter) -> Layer:
        """create parameter"""
        return rfl.make_layer(
            {"class": "variable", "shape": tensor.dims, "dtype": tensor.dtype}, name=tensor.name, out=tensor
        ).raw_tensor

    @staticmethod
    def set_parameter_initial_value(param: rf.Parameter[Layer], value: Union[None, Tensor, rf.RawTensorTypes]) -> None:
        """set parameter initial value"""
        if value is None:
            param.raw_tensor.layer_dict.pop("init", None)
            param.raw_tensor.layer_dict.pop("init_by_layer", None)
        elif isinstance(value, Tensor):
            param.raw_tensor.layer_dict.pop("init", None)
            if not value.raw_tensor.parent.can_access_children_from_root:
                accessible_parent = value.raw_tensor.parent
                while not accessible_parent.can_access_children_from_root:
                    accessible_parent = accessible_parent.parent
                value.raw_tensor.assign_parent(accessible_parent)
                # We could also maybe move out all the dependencies.
                # However, it's not clear whether this is always safe.
                for dep in value.raw_tensor.get_tensor_dependencies():
                    assert (
                        dep.parent.can_access_children_from_root
                    ), f"dep {dep} of moved value {value} is not accessible"
            param.raw_tensor.layer_dict["init_by_layer"] = value
        else:
            param.raw_tensor.layer_dict.pop("init_by_layer", None)
            param.raw_tensor.layer_dict["init"] = value
        if rfl.is_debug_eager_mode_enabled():
            shape = [d.get_dim_value() for d in param.dims]
            if isinstance(value, Tensor):
                assert value.placeholder is not None
                value_tf = value.placeholder
            else:
                value_tf = tf.broadcast_to(tf.convert_to_tensor(value), shape)
            if param.raw_tensor.debug_layer.output.placeholder is None:
                var = tf.Variable(value_tf, shape=[d.get_dim_value() for d in param.dims], dtype=param.dtype)
                param.raw_tensor.debug_layer.output.placeholder = var
            else:
                var = param.raw_tensor.debug_layer.output.placeholder
                assert isinstance(var, tf.Variable)
                var.assign(value_tf)

    @staticmethod
    def set_parameter_trainable(param: rf.Parameter, trainable: bool) -> None:
        """set parameter trainable"""
        if trainable:
            # pop it, as it's the default
            param.raw_tensor.layer_dict.pop("trainable", None)
        else:
            param.raw_tensor.layer_dict["trainable"] = False

    @staticmethod
    def convert_to_tensor(
        value: Union[Tensor, Layer, RawTensorTypes],
        *,
        dims: Sequence[Dim],
        dtype: str,
        sparse_dim: Optional[Dim] = None,
        name: Optional[str] = None,
    ) -> Tensor[Layer]:
        """convert to tensor"""
        if isinstance(value, Tensor):
            return value
        kwargs = {}
        if sparse_dim:
            kwargs["sparse_dim"] = sparse_dim
        dim_deps = _dims.get_dim_deps(dims)
        if dim_deps:
            kwargs["shape_deps"] = dim_deps
        return rfl.make_layer(
            {"class": "constant", "value": value, "shape": dims, "dtype": dtype, **kwargs}, name=name or "constant"
        )

    @classmethod
    def compare(
        cls,
        a: Union[Tensor, RawTensorTypes],
        kind: str,
        b: Union[Tensor, RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """compare"""
        kwargs = {}
        if allow_broadcast_all_sources is not None:
            kwargs["allow_broadcast_all_sources"] = allow_broadcast_all_sources
        a = rf.convert_to_tensor(a, _backend=cls)
        b = rf.convert_to_tensor(b, _backend=cls)
        return rfl.make_layer({"class": "compare", "from": [a, b], "kind": kind, **kwargs}, name=kind)

    @classmethod
    def combine(
        cls,
        a: Union[Tensor, RawTensorTypes],
        kind: str,
        b: Union[Tensor, RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """combine"""
        kwargs = {}
        if allow_broadcast_all_sources is not None:
            kwargs["allow_broadcast_all_sources"] = allow_broadcast_all_sources
        a = rf.convert_to_tensor(a, _backend=cls)
        b = rf.convert_to_tensor(b, _backend=cls)
        return rfl.make_layer({"class": "combine", "from": [a, b], "kind": kind, **kwargs}, name=kind)

    @staticmethod
    def gather(
        source: Tensor,
        *,
        indices: Union[Tensor, int],
        axis: Dim,
        clip_to_valid: bool = False,
    ) -> Tensor:
        """gather"""
        args = {}
        if clip_to_valid:
            args["clip_to_valid"] = clip_to_valid
        return rfl.make_layer(
            {"class": "gather", "from": source, "position": indices, "axis": axis, **args}, name="gather"
        )

    @staticmethod
    def slice(
        source: Tensor,
        *,
        axis: Dim,
        start: Optional[Union[int, Tensor]] = None,
        end: Optional[Union[int, Tensor]] = None,
        step: Optional[Union[int, Tensor]] = None,
        size: Optional[Union[int, Tensor, Dim]] = None,
        out_dim: Dim,
    ) -> Tensor:
        """slice"""
        if size is not None:
            assert end is None  # not implemented
            assert step is None  # not implemented
            assert size is not None  # not implemented
            return rfl.make_layer(
                {
                    "class": "slice_nd",
                    "from": source,
                    "axis": axis,
                    "start": start,
                    "size": size,
                    "out_spatial_dim": out_dim,
                },
                name="slice_nd",
            )
        assert size is None  # not implemented
        assert isinstance(start, (int, type(None)))  # not implemented
        assert isinstance(end, (int, type(None)))  # not implemented
        assert isinstance(step, (int, type(None)))  # not implemented
        args = {}
        if start is not None:
            args["slice_start"] = start
        if end is not None:
            args["slice_end"] = end
        if step is not None:
            args["slice_step"] = step
        return rfl.make_layer(
            {"class": "slice", "from": source, "axis": axis, "out_dim": out_dim, **args}, name="slice"
        )

    @staticmethod
    def matmul(a: Tensor, b: Tensor, *, reduce: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor:
        """matmul"""
        args = {}
        if not use_mask:
            args["use_mask"] = False
        return rfl.make_layer({"class": "dot", "from": [a, b], "reduce": reduce, **args}, name="matmul")

    @staticmethod
    def range_over_dim(dim: Dim, *, dtype: Optional[str] = None) -> Tensor:
        """range over dim"""
        if not dtype and dim.dyn_size_ext:
            dtype = dim.dyn_size_ext.dtype
        if not dtype:
            dtype = rf.get_default_array_index_dtype()
        return rfl.make_layer(
            {"class": "range_in_axis", "from": _dims.get_dim_deps(dim), "axis": dim, "dtype": dtype},
            name="range_over_dim",
        )

    @staticmethod
    def replace_dim(source: Tensor, *, in_dim: Dim, out_dim: Dim) -> Tensor:
        """
        :param source:
        :param in_dim:
        :param out_dim:
        :return: source with in_dim replaced by out_dim.
        """
        return rfl.make_layer(
            {"class": "reinterpret_data", "set_dim_tags": {in_dim: out_dim}, "from": source}, name="new_dim"
        )

    @staticmethod
    def reduce(source: Tensor, *, mode: str, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor:
        """Reduce"""
        assert mode in Backend._AllowedReduceModes
        kwargs = {}
        if not use_mask:
            kwargs["use_time_mask"] = False
        return rfl.make_layer(
            {"class": "reduce", "from": source, "mode": mode, "axis": axis, **kwargs}, name=f"reduce_{mode}"
        )

    @staticmethod
    @contextlib.contextmanager
    def random_journal_replay(journal: Sequence[Dict[str, Any]]):
        """
        Replays the journal.
        At exit, the journal is cleared, and we check that we replayed everything.
        """
        try:
            ReturnnLayersBackend._random_journal_replay_enabled = True
            ReturnnLayersBackend._random_journal_replay_idx = 0
            ReturnnLayersBackend._random_journal = journal
            yield
        finally:
            final_replay_idx = ReturnnLayersBackend._random_journal_replay_idx
            ReturnnLayersBackend._random_journal_replay_enabled = False
            ReturnnLayersBackend._random_journal_replay_idx = 0
            ReturnnLayersBackend._random_journal = None
        assert final_replay_idx == len(journal)

    _random_journal_replay_enabled = False
    _random_journal_replay_idx = 0
    _random_journal = None  # type: Optional[Sequence[Dict[str, Any]]]

    @staticmethod
    def random(
        *,
        dims: Sequence[Dim],
        dtype: str,
        sparse_dim: Optional[Dim] = None,
        distribution: str,
        mean: Optional[Union[int, float, Tensor]] = None,
        stddev: Optional[Union[int, float, Tensor]] = None,
        bound: Optional[Union[int, float, Tensor]] = None,
        minval: Optional[Union[int, float, Tensor]] = None,
        maxval: Optional[Union[int, float, Tensor]] = None,
        seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = None,
        algorithm: Optional[str] = None,
        explicit_state: Optional[Tensor] = None,
        auto_update_state: Optional[bool] = None,
        static: Optional[bool] = None,
        out: Optional[Tensor] = None,
    ) -> Tensor:
        """random"""
        if ReturnnLayersBackend._random_journal_replay_enabled:
            idx = ReturnnLayersBackend._random_journal_replay_idx
            ReturnnLayersBackend._random_journal_replay_idx += 1
            elem = ReturnnLayersBackend._random_journal[idx]
            assert isinstance(elem, dict)
            # Different calls might create their own dims, so we cannot directly compare dim tags for equality.
            assert len(elem["dims"]) == len(dims)
            # We still check static dimension equality.
            for dim_journal, dim_tf in zip(elem["dims"], dims):
                assert (
                    dim_journal.dimension == dim_tf.dimension
                ), f"random with dims {dims} does not match random journal replay\n{elem!r}"
            assert elem["dtype"] == dtype
            assert elem["sparse_dim"] == sparse_dim
            assert elem["distribution"] == distribution
            assert rfl.Layer.inner_control_flow() is None  # not implemented yet
            return rfl.make_layer(
                {
                    "class": "eval",
                    "from": (),
                    "eval": _random_replay_eval,
                    "eval_locals": {"idx": idx},
                    "out_type": {"dims": dims, "dtype": dtype, "sparse_dim": sparse_dim},
                },
                name="random_replay",
            )
        kwargs = {
            "mean": mean,
            "stddev": stddev,
            "bound": bound,
            "minval": minval,
            "maxval": maxval,
            "seed": seed,
            "algorithm": algorithm,
            "explicit_state": explicit_state,
            "auto_update_state": auto_update_state,
            "static": static,
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return rfl.make_layer(
            {
                "class": "random",
                "shape": dims,
                "shape_deps": _dims.get_dim_deps(dims),
                "dtype": dtype,
                "sparse_dim": sparse_dim,
                "distribution": distribution,
                "stop_grad": True,
                **kwargs,
            },
            name="random",
        )

    @staticmethod
    def masked_select(
        tensor: Tensor, *, mask: Tensor, dims: Sequence[Dim], out_dim: Optional[Dim] = None
    ) -> Tuple[Tensor, Dim]:
        """
        :param tensor:
        :param mask:
        :param dims: the order of the dims defines the format. those dims should be exactly the dims of the mask.
        :param out_dim:
        :return: tensor where all dims in mask/dims are removed and replaced by a new dim.
            the new dim is also returned.
            if mask==True for all elements, the returned tensor would be simply the flattened input tensor.
        """
        assert mask.dtype == "bool"
        assert set(mask.dims) == set(dims)
        assert set(mask.dims).issubset(set(tensor.dims))
        if not out_dim:
            out_dim = Dim(None, name="mask")
        return (
            rfl.make_layer(
                {"class": "boolean_mask", "from": tensor, "mask": mask, "dims": dims, "out_dim": out_dim},
                name="boolean_mask",
            ),
            out_dim,
        )

    @staticmethod
    def batch_norm(
        source: Tensor,
        *,
        in_dim: Union[Dim, Sequence[Dim]],
        running_mean: Tensor,
        running_variance: Tensor,
        gamma: Optional[Tensor],
        beta: Optional[Tensor],
        epsilon: float,
        momentum: float,
        affine: bool,
        use_mask: bool,
    ) -> Tensor:
        """batch norm"""
        reuse_params = {
            "batch_norm/v2_mean": running_mean,
            "batch_norm/v2_variance": running_variance,
        }
        if affine:
            reuse_params["batch_norm/v2_gamma"] = gamma
            reuse_params["batch_norm/v2_beta"] = beta
        reuse_params = {"map": {k: {"layer_output": v} for k, v in reuse_params.items()}}
        return rfl.make_layer(
            {
                "class": "batch_norm",
                "from": source,
                "in_dim": in_dim,
                "use_std": affine,
                "use_shift": affine,
                "param_version": 2,
                "reuse_params": reuse_params,
                "momentum": momentum,
                "epsilon": epsilon,
                "masked_time": use_mask,
            },
            name="batch_norm",
        )

    # noinspection PyShadowingBuiltins
    @staticmethod
    def conv(
        source: Tensor,
        *,
        in_dim: Dim,
        out_dim: Dim,
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
        filter: Tensor,
        filter_size: Sequence[Dim],  # to have the order well-defined
        padding: str,
        strides: Optional[Union[int, Sequence[int]]] = None,
        dilation_rate: Optional[Union[int, Sequence[int]]] = None,
        groups: Optional[int] = None,
        bias: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """conv"""
        if not out_spatial_dims:
            out_spatial_dims = rf.make_conv_out_spatial_dims(
                description_prefix="conv",
                in_spatial_dims=in_spatial_dims,
                filter_size=[d.dimension for d in filter_size],
                strides=strides or 1,
                dilation_rate=dilation_rate or 1,
                padding=padding,
            )
        layer_dict = {
            "class": "conv",
            "from": source,
            "in_dim": in_dim,
            "in_spatial_dims": in_spatial_dims,
            "out_dim": out_dim,
            "out_spatial_dims": out_spatial_dims,
            "filter_size": filter_size,
            "padding": padding,
        }
        if strides:
            layer_dict["strides"] = strides
        if dilation_rate:
            layer_dict["dilation_rate"] = dilation_rate
        if groups:
            layer_dict["groups"] = groups
        layer_dict.update({"filter": filter, "with_bias": bias is not None})
        if bias is not None:
            layer_dict["bias"] = bias
        out = rfl.make_layer(layer_dict, name="conv")
        return out, out_spatial_dims

    # noinspection PyShadowingBuiltins
    @staticmethod
    def transposed_conv(
        source: Tensor,
        *,
        in_dim: Dim,
        out_dim: Dim,
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
        filter: Tensor,
        filter_size: Sequence[Dim],
        padding: str,
        remove_padding: Union[Sequence[int], int] = 0,
        output_padding: Optional[Union[Sequence[Optional[int]], int]] = None,
        strides: Optional[Sequence[int]] = None,
        bias: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """transposed conv"""
        if not out_spatial_dims:
            out_spatial_dims = [Dim(None, name=f"out-spatial-dim{i}") for i, s in enumerate(filter_size)]
            for i in range(len(filter_size)):
                s = filter_size[i].dimension if not strides else strides[i]
                if filter_size[i].dimension == s == 1 or (s == 1 and padding.lower() == "same"):
                    out_spatial_dims[i] = in_spatial_dims[i]
        layer_dict = {
            "class": "transposed_conv",
            "from": source,
            "in_dim": in_dim,
            "in_spatial_dims": in_spatial_dims,
            "out_dim": out_dim,
            "out_spatial_dims": out_spatial_dims,
            "filter_size": filter_size,
            "padding": padding,
        }
        if remove_padding:
            layer_dict["remove_padding"] = remove_padding
        if output_padding:
            layer_dict["output_padding"] = output_padding
        if strides:
            layer_dict["strides"] = strides
        layer_dict.update({"filter": filter, "with_bias": bias is not None})
        if bias is not None:
            layer_dict["bias"] = bias
        out = rfl.make_layer(layer_dict, name="transposed_conv")
        return out, out_spatial_dims

    @staticmethod
    def pool(
        source: Tensor,
        *,
        mode: str,
        pool_size: Sequence[int],
        padding: str = "valid",
        dilation_rate: Union[Sequence[int], int] = 1,
        strides: Sequence[int],
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """pool"""
        if out_spatial_dims is None:
            out_spatial_dims = rf.make_conv_out_spatial_dims(
                description_prefix="pool",
                in_spatial_dims=in_spatial_dims,
                filter_size=pool_size,
                strides=strides,
                dilation_rate=dilation_rate,
                padding=padding,
            )
        other_dims = [d for d in source.dims if d not in in_spatial_dims and not d.is_batch_dim()]
        assert other_dims  # currently not implemented otherwise, need in_dim...
        if source.feature_dim and source.feature_dim in other_dims:
            in_dim = source.feature_dim
        else:
            in_dim = other_dims[-1]
        args = {
            "mode": mode,
            "pool_size": pool_size,
            "padding": padding,
            "dilation_rate": dilation_rate,
            "strides": strides,
            "in_spatial_dims": in_spatial_dims,
            "out_spatial_dims": out_spatial_dims,
            "in_dim": in_dim,  # it does not really matter, but we need sth currently
        }
        layer = rfl.make_layer({"class": "pool", "from": source, **args}, name="pool")
        if source.feature_dim != in_dim:
            # We want that the feature-dim stays consistent. PoolLayer currently just sets it to the in_dim.
            layer = rfl.make_layer(
                {"class": "reinterpret_data", "from": layer, "set_axes": {"F": source.feature_dim}},
                name="pool_reset_feature",
            )
        return layer, out_spatial_dims


def _random_replay_eval(idx, **_kwargs):
    # noinspection PyProtectedMember
    elem = ReturnnLayersBackend._random_journal[idx]
    assert isinstance(elem, dict)
    out = elem["out"]
    assert isinstance(out, Tensor)
    assert isinstance(out.raw_tensor, numpy.ndarray)
    return tf.constant(out.raw_tensor, dtype=out.dtype)
